
from typing import Dict, Type

from py2puml.domain.umlclass import UmlClass, UmlAttribute
from py2puml.domain.umlitem import UmlItem


def inspect_namedtuple_type(
    namedtuple_type: Type,
    namedtuple_type_fqn: str,
    domain_items_by_fqn: Dict[str, UmlItem]
):
    domain_items_by_fqn[namedtuple_type_fqn] = UmlClass(
        name=namedtuple_type.__name__,
        fqn=namedtuple_type_fqn,
        attributes=[
            UmlAttribute(tuple_field, 'Any', False)
            for tuple_field in namedtuple_type._fields
        ]
    )




from re import compile, Pattern
from typing import List, Tuple

IS_COMPOUND_TYPE: Pattern = compile('^[a-z|A-Z|\\[|\\]|\\.|,|\\s|_]+$')
SPLITTING_CHARACTERS = ('[', ']', ',')

class CompoundTypeSplitter(object):
    def __init__(self, compound_type_annotation: str):
        if (compound_type_annotation is None) or not IS_COMPOUND_TYPE.match(compound_type_annotation):
            raise ValueError(f'{compound_type_annotation} seems to be an invalid type annotation')

        self.compound_type_annotation = compound_type_annotation
    
    def get_parts(self) -> Tuple[str]:
        parts = [self.compound_type_annotation]
        for splitting_character in SPLITTING_CHARACTERS:
            new_parts = []
            for part in parts:
                splitted_parts = part.split(splitting_character)
                new_parts.append(splitted_parts[0])
                if len(splitted_parts) > 1:
                    for splitted_part in splitted_parts[1:]:
                        new_parts.extend([splitting_character, splitted_part])
            parts = [
                new_part.strip()
                for new_part in new_parts
                if len(new_part.strip()) > 0
            ]
        return tuple(parts)




from typing import Dict, Iterable, List, Type
from types import ModuleType

from dataclasses import is_dataclass
from enum import Enum
from inspect import getmembers, isclass

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.inspection.inspectclass import inspect_dataclass_type, inspect_class_type
from py2puml.inspection.inspectenum import inspect_enum_type
from py2puml.inspection.inspectnamedtuple import inspect_namedtuple_type

def filter_domain_definitions(module: ModuleType, root_module_name: str) -> Iterable[Type]:
    for definition_key in dir(module):
        definition_type = getattr(module, definition_key)
        if isclass(definition_type):
            definition_members = getmembers(definition_type)
            definition_module_member = next((
                member for member in definition_members
                # ensures that the type belongs to the module being parsed
                if member[0] == '__module__' and member[1].startswith(root_module_name)
            ), None)
            if definition_module_member is not None:
                yield definition_type

def inspect_domain_definition(
    definition_type: Type,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    definition_type_fqn = f'{definition_type.__module__}.{definition_type.__name__}'
    if definition_type_fqn not in domain_items_by_fqn:
        if issubclass(definition_type, Enum):
            inspect_enum_type(definition_type, definition_type_fqn, domain_items_by_fqn)
        elif getattr(definition_type, '_fields', None) is not None:
            inspect_namedtuple_type(definition_type, definition_type_fqn, domain_items_by_fqn)
        elif is_dataclass(definition_type):
            inspect_dataclass_type(
                definition_type, definition_type_fqn,
                root_module_name, domain_items_by_fqn, domain_relations
            )
        else:
            inspect_class_type(
                definition_type, definition_type_fqn,
                root_module_name, domain_items_by_fqn, domain_relations
            )

def inspect_module(
    domain_item_module: ModuleType,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    # processes only the definitions declared or imported within the given root module
    for definition_type in filter_domain_definitions(domain_item_module, root_module_name):
        inspect_domain_definition(definition_type, root_module_name, domain_items_by_fqn, domain_relations)




from typing import Type, List, Dict

from re import compile
from dataclasses import dataclass

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlclass import UmlClass, UmlAttribute
from py2puml.domain.umlrelation import UmlRelation, RelType
from py2puml.parsing.parseclassconstructor import parse_class_constructor
from py2puml.utils import inspect_domain_definition

CONCRETE_TYPE_PATTERN = compile("^<(?:class|enum) '([\\.|\\w]+)'>$")

def get_type_name(type: Type, root_module_name: str):
    if type.__module__.startswith(root_module_name):
        return type.__name__
    else:
        return f'{type.__module__}.{type.__name__}'

def handle_inheritance_relation(
    class_type: Type,
    class_fqn: str,
    root_module_name: str,
    domain_relations: List[UmlRelation]
):
    for base_type in getattr(class_type, '__bases__', ()):
        base_type_fqn = f'{base_type.__module__}.{base_type.__name__}'
        if base_type_fqn.startswith(root_module_name):
            domain_relations.append(
                UmlRelation(base_type_fqn, class_fqn, RelType.INHERITANCE)
            )

def inspect_static_attributes(
    class_type: Type,
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
) -> List[UmlAttribute]:
    definition_attrs: List[UmlAttribute] = []
    uml_class = UmlClass(
        name=class_type.__name__,
        fqn=class_type_fqn,
        attributes=definition_attrs
    )
    domain_items_by_fqn[class_type_fqn] = uml_class
    # inspect_domain_definition(class_type)
    type_annotations = getattr(class_type, '__annotations__', None)
    if type_annotations is not None:
        for attr_name, attr_class in type_annotations.items():
            attr_raw_type = str(attr_class)
            concrete_type_match = CONCRETE_TYPE_PATTERN.search(attr_raw_type)
            if concrete_type_match:
                concrete_type = concrete_type_match.group(1)
                if attr_class.__module__.startswith(root_module_name):
                    attr_type = attr_class.__name__
                    domain_relations.append(
                        UmlRelation(uml_class.fqn, f'{attr_class.__module__}.{attr_class.__name__}', RelType.COMPOSITION)
                    )
                else:
                    attr_type = concrete_type
            else:
                composition_rel = getattr(attr_class, '_name', None)
                component_classes = getattr(attr_class, '__args__', None)
                if composition_rel and component_classes:
                    component_names = [
                        get_type_name(component_class, root_module_name)
                        for component_class in component_classes
                        # filters out forward refs
                        if getattr(component_class, '__name__', None) is not None
                    ]
                    domain_relations.extend([
                        UmlRelation(uml_class.fqn, f'{component_class.__module__}.{component_class.__name__}', RelType.COMPOSITION)
                        for component_class in component_classes
                        if component_class.__module__.startswith(root_module_name)
                    ])
                    attr_type = f"{composition_rel}[{', '.join(component_names)}]"
                else:
                    attr_type = attr_raw_type
            uml_attr = UmlAttribute(attr_name, attr_type, static=True)
            definition_attrs.append(uml_attr)

    return definition_attrs

def inspect_class_type(
    class_type: Type,
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    attributes = inspect_static_attributes(
        class_type, class_type_fqn, root_module_name,
        domain_items_by_fqn, domain_relations
    )
    instance_attributes, compositions = parse_class_constructor(class_type, class_type_fqn, root_module_name)
    attributes.extend(instance_attributes)
    domain_relations.extend(compositions.values())

    handle_inheritance_relation(class_type, class_type_fqn, root_module_name, domain_relations)

def inspect_dataclass_type(
    class_type: Type[dataclass],
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation]
):
    for attribute in inspect_static_attributes(
        class_type,
        class_type_fqn,
        root_module_name,
        domain_items_by_fqn,
        domain_relations
    ):
        attribute.static = False

    handle_inheritance_relation(class_type, class_type_fqn, root_module_name, domain_relations)



from typing import Dict, Type
from enum import Enum

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlenum import UmlEnum, Member

def inspect_enum_type(
    enum_type: Type[Enum],
    enum_type_fqn: str,
    domain_items_by_fqn: Dict[str, UmlItem]
):
    domain_items_by_fqn[enum_type_fqn] = UmlEnum(
        name=enum_type.__name__,
        fqn=enum_type_fqn,
        members=[
            Member(name=enum_member.name, value=enum_member.value)
            for enum_member in enum_type
        ]
    )




from typing import Dict, List, Tuple, Type

from ast import parse, AST
from importlib import import_module
from inspect import getsource, unwrap
from textwrap import dedent

from py2puml.domain.umlclass import UmlAttribute
from py2puml.domain.umlrelation import UmlRelation
from py2puml.parsing.astvisitors import ConstructorVisitor
from py2puml.parsing.moduleresolver import ModuleResolver

def parse_class_constructor(
    class_type: Type,
    class_fqn: str,
    root_module_name: str
) -> Tuple[List[UmlAttribute], Dict[str, UmlRelation]]:
    constructor = getattr(class_type, '__init__', None)
    # conditions to meet in order to parse the AST of a constructor 
    if (
        # the constructor must be defined
        constructor is None
    ) or (
        # the constructor's source code must be available
        not hasattr(constructor, '__code__')
    ) or (
        # the constructor must belong to the parsed class (not its parent's one)
        not constructor.__qualname__.endswith(f'{class_type.__name__}.__init__')
    ):
        return [], {}

    # gets the original constructor, if wrapped by a decorator
    constructor = unwrap(constructor)

    constructor_source: str = dedent(getsource(constructor.__code__))
    constructor_ast: AST = parse(constructor_source)

    module_resolver = ModuleResolver(import_module(class_type.__module__))

    visitor = ConstructorVisitor(constructor_source, class_type.__name__, root_module_name, module_resolver)
    visitor.visit(constructor_ast)

    return visitor.uml_attributes, visitor.uml_relations_by_target_fqn



from typing import Dict, List, Iterable
from types import ModuleType
from pkgutil import walk_packages
from importlib import import_module

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.inspection.inspectmodule import inspect_module
from py2puml.exportpuml import to_puml_content


def py2puml(domain_path: str, domain_module: str) -> Iterable[str]:
    domain_items_by_fqn: Dict[str, UmlItem] = {}
    domain_relations: List[UmlRelation] = []
    for _, name, is_pkg in walk_packages([domain_path], f'{domain_module}.'):
        if not is_pkg:
            domain_item_module: ModuleType = import_module(name)
            inspect_module(
                domain_item_module,
                domain_module,
                domain_items_by_fqn,
                domain_relations
            )
    return to_puml_content(domain_items_by_fqn.values(), domain_relations)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py2puml', 'py2puml.domain', 'py2puml.inspection', 'py2puml.parsing']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['py2puml = py2puml.cli:run']}

setup_kwargs = {
    'name': 'py2puml',
    'version': '0.5.3',
    'description': 'Generate PlantUML class diagrams to document your Python application.',
    'long_description': '<div align="center">\n  <a href="https://www.python.org/psf-landing/" target="_blank">\n    <img width="350px" alt="Python logo"\n      src="https://www.python.org/static/community_logos/python-logo-generic.svg" />\n  </a>\n  <a href="http://plantuml.com/" target="_blank">\n    <img width="116px" height="112px" alt="PlantUML logo" src="http://s.plantuml.com/logoc.png" style="margin-bottom: 40px" vspace="40px" />\n  </a>\n  <h1>Python to PlantUML</h1>\n</div>\n\nGenerate PlantUML class diagrams to document your Python application.\n\n# How it works\n\n`py2puml` produces a class diagram [PlantUML script](https://plantuml.com/en/class-diagram) representing classes properties (static and instance attributes) and their relations (composition and inheritance relationships).\n\n`py2puml` internally uses code [inspection](https://docs.python.org/3/library/inspect.html) (also called *reflexion* in other programming languages) and [abstract tree parsing](https://docs.python.org/3/library/ast.html) to retrieve relevant information.\nSome parsing features are available only since Python 3.8 (like [ast.get_source_segment](https://docs.python.org/3/library/ast.html#ast.get_source_segment)).\n\n## Features\n\nFrom a given path corresponding to a folder containing Python code, `py2puml` processes each file as a module and generates a [PlantUML script](https://plantuml.com/en/class-diagram) of its classe-like definitions using:\n\n* **[inspection](https://docs.python.org/3/library/inspect.html)** and [type annotations](https://docs.python.org/3/library/typing.html) to detect:\n  * static class attributes and [dataclass](https://docs.python.org/3/library/dataclasses.html) fields\n  * fields of [namedtuples](https://docs.python.org/3/library/collections.html#collections.namedtuple)\n  * members of [enumerations](https://docs.python.org/3/library/enum.html)\n  * composition and inheritance relationships (between your domain classes only, for documentation sake).\nThe detection of composition relationships relies on type annotations only, assigned values or expressions are never evaluated to prevent unwanted side-effects\n\n* parsing [abstract syntax trees](https://docs.python.org/3/library/ast.html#ast.NodeVisitor) to detect the instance attributes defined in `__init__` constructors\n\n`py2puml` outputs diagrams in PlantUML syntax, which can be:\n* versioned along your code with a unit-test ensuring its consistency (see the [test_py2puml.py\'s test_py2puml_model_on_py2uml_domain](tests/py2puml/test_py2puml.py) example)\n* generated and hosted along other code documentation (better option: generated documentation should not be versioned with the codebase)\n\nTo generate image files, use the PlantUML runtime, a docker image of the runtime (see [think/plantuml](https://hub.docker.com/r/think/plantuml)) or of a server (see the CLI documentation below)\n\nIf you like tools related with PlantUML, you may also be interested in this [lucsorel/plantuml-file-loader](https://github.com/lucsorel/plantuml-file-loader) project:\na webpack loader which converts PlantUML files into images during the webpack processing (useful to [include PlantUML diagrams in your slides](https://github.com/lucsorel/markdown-image-loader/blob/master/README.md#web-based-slideshows) with RevealJS or RemarkJS).\n\n# Install\n\nInstall from [PyPI](https://pypi.org/project/py2puml/):\n\n* with `pip`:\n\n```sh\npip install py2puml\n```\n\n* with [poetry](https://python-poetry.org/docs/):\n\n```sh\npoetry add py2puml\n```\n\n* with [pipenv](https://pipenv.readthedocs.io/en/latest/):\n\n```sh\npipenv install py2puml\n```\n\n# Usage\n\n## CLI\n\nOnce `py2puml` is installed at the system level, an eponymous command is available in your environment shell.\n\nFor example, to create the diagram of the classes used by `py2puml`, run:\n\n```sh\npy2puml py2puml/domain py2puml.domain\n```\n\nThis outputs the following PlantUML script:\n\n```plantuml\n@startuml\nclass py2puml.domain.umlclass.UmlAttribute {\n  name: str\n  type: str\n  static: bool\n}\nclass py2puml.domain.umlclass.UmlClass {\n  attributes: List[UmlAttribute]\n}\nclass py2puml.domain.umlitem.UmlItem {\n  name: str\n  fqn: str\n}\nclass py2puml.domain.umlenum.Member {\n  name: str\n  value: str\n}\nclass py2puml.domain.umlenum.UmlEnum {\n  members: List[Member]\n}\nenum py2puml.domain.umlrelation.RelType {\n  COMPOSITION: * {static}\n  INHERITANCE: <| {static}\n}\nclass py2puml.domain.umlrelation.UmlRelation {\n  source_fqn: str\n  target_fqn: str\n  type: RelType\n}\npy2puml.domain.umlclass.UmlClass *-- py2puml.domain.umlclass.UmlAttribute\npy2puml.domain.umlitem.UmlItem <|-- py2puml.domain.umlclass.UmlClass\npy2puml.domain.umlenum.UmlEnum *-- py2puml.domain.umlenum.Member\npy2puml.domain.umlitem.UmlItem <|-- py2puml.domain.umlenum.UmlEnum\npy2puml.domain.umlrelation.UmlRelation *-- py2puml.domain.umlrelation.RelType\n@enduml\n```\n\nUsing PlantUML, this script renders this diagram:\n\n![py2puml UML Diagram](http://www.plantuml.com/plantuml/png/ZPB1Qy8m5CRl-IlUMR277Oi7HOGLfXrTTnfZfFLj59AqIru7elxlUuqQ1ewcftmcvlTzUL-NZgIbNYjHA-aST8U7Zdyb-rRBnYGi_NxogjMAo3PLJmX70M2anXGSMTPqw89c7ZLr2bNRAd6EKzU3y4HvuxiKdXf7RtyztqTO3Q4UK1clTza-lusN8_VHz3hPegxGtbt_aQh7IG0EiE7L4xI7tTvnGGyl6FxuptsBYeVMcgH0LV8iFMETRv_pbwpCybqACpXU1dlcasptk2coShLRRr9OdC9HI3ZYm2cBg_OkhkrjZHzXIW0axHTIs0drNhEnIRJDsNm-wKCIaIuN9mR5t4Ia3muptlca5ECcOjh4VPPu_MA9Pi_xlm00)\n\nFor a full overview of the CLI, run:\n\n```sh\npy2puml --help\n```\n\nThe CLI can also be launched as a python module:\n\n```sh\npython -m py2puml py2puml/domain py2puml.domain\n```\n\nPipe the result of the CLI with a PlantUML server for instantaneous documentation (rendered by ImageMagick):\n\n```sh\n# runs a local PlantUML server from a docker container:\ndocker run -d --rm -p 1234:8080 --name plantumlserver plantuml/plantuml-server:jetty \n\npy2puml py2puml/domain py2puml.domain | curl -X POST --data-binary @- http://localhost:1234/svg/ --output - | display\n\n# stops the container when you don\'t need it anymore, restarts it later\ndocker stop plantumlserver\ndocker start plantumlserver\n```\n\n## Python API\n\nFor example, to create the diagram of the classes used by `py2puml`:\n\n* import the py2puml function in your script (see [py2puml/example.py](py2puml/example.py)):\n\n```python\nfrom py2puml.py2puml import py2puml\n\n# outputs the PlantUML content in the terminal\nprint(\'\'.join(py2puml(\'py2puml/domain\', \'py2puml.domain\')))\n\n# writes the PlantUML content in a file\nwith open(\'py2puml/domain.puml\', \'w\') as puml_file:\n    puml_file.writelines(py2puml(\'py2puml/domain\', \'py2puml.domain\'))\n```\n\n* running it (`python3 -m py2puml.example`) will output the previous PlantUML diagram in the terminal and write it in a file.\n\n\n# Tests\n\n```sh\n# directly with poetry\npoetry run python -m pytest -v\n\n# in a virtual environment\npython3 -m pytest -v\n```\n\nCode coverage (with missed [branch statements](https://pytest-cov.readthedocs.io/en/latest/config.html?highlight=--cov-branch)):\n\n```sh\npoetry run python -m pytest -v --cov=py2puml --cov-branch --cov-report term-missing --cov-fail-under 90\n```\n\n# Changelog\n\n* `0.5.3`: handle constructors decorated by wrapping decorators (decorators making uses of `functools.wrap`)\n* `0.5.2`: specify in pyproject.toml that py2puml requires python 3.8+ (`ast.get_source_segment` was introduced in 3.8)\n* `0.5.1`: prevent from parsing inherited constructors\n* `0.5.0`: handle instance attributes in class constructors, add code coverage of unit tests\n* `0.4.0`: add a simple CLI\n* `0.3.1`: inspect sub-folders recursively\n* `0.3.0`: handle classes derived from namedtuples (attribute types are `Any`)\n* `0.2.0`: handle inheritance relationships and enums\n* `0.1.3`: first release, handle all modules of a folder and compositions of domain classes\n\n# Licence\n\nUnless stated otherwise all works are licensed under the [MIT license](http://spdx.org/licenses/MIT.html), a copy of which is included [here](LICENSE).\n\n# Contributions\n\n* [Luc Sorel-Giffo](https://github.com/lucsorel)\n* [Doyou Jung](https://github.com/doyou89)\n* [Julien Jerphanion](https://github.com/jjerphan)\n\nPull-requests are welcome and will be processed on a best-effort basis.\nFollow the [contributing guide](CONTRIBUTING.md).\n\n# Current limitations\n\n* regarding **inspection**\n\n  * type hinting is optional when writing Python code and discarded when it is executed, as mentionned in the [typing official documentation](https://docs.python.org/3/library/typing.html). The quality of the diagram output by `py2puml` depends on the reliability with which the type annotations were written\n\n  > The Python runtime does not enforce function and variable type annotations. They can be used by third party tools such as type checkers, IDEs, linters, etc.\n\n  * complex type hints with more than one level of genericity are not properly handled for the moment: `List[MyClass]` or `Dict[str, MyClass]` are handled properly, `Dict[str, List[MyClass]]` is not.\n  If your domain classes (also called business objects or DTOs) have attributes with complex type hints, it may be a code smell indicating that you should write a class which would better represent the business logic.\n  But I may improve this part of the library as well 😀\n\n* regarding the detection of instance attributes with **AST parsing**:\n  * only constructors are visited, attributes assigned in other functions won\'t be documented\n  * attribute types are inferred from type annotations:\n    * of the attribute itself\n    * of the variable assigned to the attribute: a signature parameter or a locale variable\n    * to avoid side-effects, no code is executed nor interpreted\n\n# Alternatives\n\nIf `py2puml` does not meet your needs (suggestions and pull-requests are **welcome**), you can have a look at these projects which follow other approaches (AST, linting, modeling):\n\n* [cb109/pyplantuml](https://github.com/cb109/pyplantuml)\n* [deadbok/py-puml-tools](https://github.com/deadbok/py-puml-tools)\n* [caballero/genUML](https://github.com/jose-caballero/genUML)\n',
    'author': 'Luc Sorel-Giffo',
    'author_email': None,
    'maintainer': 'Luc Sorel-Giffo',
    'maintainer_email': None,
    'url': 'https://github.com/lucsorel/py2puml',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




from typing import Type, Tuple
from types import ModuleType
from collections import namedtuple

from functools import reduce

NamespacedType = namedtuple('NamespacedType', ['full_namespace', 'type_name'])

EMPTY_NAMESPACED_TYPE = NamespacedType(None, None)

def search_in_module_or_builtins(searched_module: ModuleType, namespace: str):
    if searched_module is None:
        return None

    # searches the namespace in the module definitions and imports
    found_module_or_leaf_type = getattr(searched_module, namespace, None)
    if found_module_or_leaf_type is not None:
        return found_module_or_leaf_type

    # searches the namespace in the builtins otherwise
    if hasattr(searched_module, '__builtins__'):
        return searched_module.__builtins__.get(namespace, None)
    else:
        return None

class ModuleResolver(object):
    def __init__(self, module: ModuleType):
        self.module = module

    def resolve_full_namespace_type(self, partial_dotted_path: str) -> Tuple[str, str]:
        '''
        Returns a tuple of 2 strings:
        - the full namespaced type
        - the short named type
        '''
        if partial_dotted_path is None:
            return EMPTY_NAMESPACED_TYPE
        
        namespaces = partial_dotted_path.split('.')
        leaf_type: Type = reduce(
            search_in_module_or_builtins,
            namespaces,
            self.module
        )
        if leaf_type is None:
            return EMPTY_NAMESPACED_TYPE
        else:
            # https://bugs.python.org/issue34422#msg323772
            short_type = getattr(leaf_type, '__name__', getattr(leaf_type, '_name', None))
            return NamespacedType(
                f'{leaf_type.__module__}.{short_type}',
                short_type
            )

    def get_module_full_name(self) -> str:
        return self.module.__name__
